from typing import List
from pydantic import BaseModel, Field

class Selector(BaseModel):
    """
    Selector class for specifying items to embed and metadata.

    This class is used to define specific criteria for data processing, particularly for selecting what items should be embedded and what should be stored as metadata.

    Attributes:
    -----------
    to_embed : List[str]
        List of items to embed, specified by their identifiers or names.

    to_metadata : List[str]
        List of metadata items to include, specified by their identifiers or names.
    """

    to_embed: List[str] = Field(default_factory=list, description="List of items to embed.")
    
    to_metadata: List[str] = Field(default_factory=list, description="List of metadata items.")


from neumai.Chunkers import (
    Chunker,
    CharacterChunker,
    CustomChunker,
    RecursiveChunker
)    
from neumai.Chunkers.ChunkerEnum import ChunkerEnum

class ChunkerFactory:
    """Class that leverages the Factory pattern to get the appropriate chunker
    """
    def get_chunker(chunker_name: str, chunker_information: dict) -> Chunker:
        if chunker_information == None:
            return RecursiveChunker()
        chunker_name_enum:ChunkerEnum = ChunkerEnum.as_chunker_enum(chunker_name)
        if chunker_name_enum == ChunkerEnum.characterchunker:
            return CharacterChunker(**chunker_information)
        elif chunker_name_enum == ChunkerEnum.customchunker:
            return CustomChunker(**chunker_information)
        elif chunker_name_enum == ChunkerEnum.recursivechunker:
            return RecursiveChunker(**chunker_information)
        else:
            return RecursiveChunker(**chunker_information)


from neumai.Shared.NeumSearch import NeumSearchResult
from neumai.Shared.NeumVector import NeumVector
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from abc import ABC, abstractmethod
from typing import List
from pydantic import BaseModel
import json

class SinkConnector(ABC, BaseModel):

    @property
    @abstractmethod
    def sink_name(self) -> str:
        pass
    
    @property
    @abstractmethod
    def required_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optional_properties(self) -> List[str]:
        pass

    @abstractmethod
    def validation(self) -> bool:
        """config_validation sink setup"""

    @abstractmethod
    def store(self, vectors_to_store:List[NeumVector]) -> int:
        """Store vectors with a given service"""

    @abstractmethod
    def search(self, vector:List[float], number_of_results:int, filter:dict={}) -> List[NeumSearchResult]:
        """Search vectors for a given service"""
    
    @abstractmethod
    def delete_vectors_with_file_id(self, file_id:str ) -> bool:
        """Deletes vectors for a specific file id"""
    
    @abstractmethod
    def info(self) -> NeumSinkInfo:
        """Get information about what is stores in the sink"""

    def as_json(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['sink_name'] = self.sink_name
        json_to_return['sink_information'] = json.loads(self.json())
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['required_properties'] = self.required_properties
        json_to_return['optional_properties'] = self.optional_properties
        return json_to_return


from datetime import datetime
from typing import List, Generator, Optional
from neumai.DataConnectors.DataConnector import DataConnector
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.Shared.Selector import Selector
from pydantic import Field
from neumai.Shared.Exceptions import NeumFileException
import tempfile

class FileConnector(DataConnector):
    """
    File Connector

    Allows user to load a publicly available file.
    
    Attributes:
    -----------

    url : str
        The url poitning to the publicly available file
    selector : Optional[Selector]
        Optional selector object to define what data data should be used to generate embeddings or stored as metadata with the vector.
    
    """
    
    url: str = Field(..., description="URL required for the connector.")

    selector: Optional[Selector] = Field(Selector(to_embed=[], to_metadata=[]), description="Selector for data connector metadata")

    @property
    def connector_name(self) -> str:
        return "FileConnector"

    @property
    def required_properties(self) -> List[str]:
        return ["url"]

    @property
    def optional_properties(self) -> List[str]:
        return []

    @property
    def available_metadata(self) -> str:
        return ['url']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "CSVLoader", "JSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        availableMetadata = {'url':self.url}
        selected_metadata  = {k: availableMetadata[k] for k in self.selector.to_metadata if k in availableMetadata}
        yield CloudFile(file_identifier=self.url, metadata=selected_metadata, id=self.url)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Delta is not different, we are just getting one file. 
        yield self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to random file location
        import requests
        import os
        from urllib.parse import urlparse

        headers = {"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.182 Safari/537.36"}
        response = requests.get(cloudFile.file_identifier, headers=headers)
        if not response.ok:
            raise NeumFileException(f"File can't be accessed. Please make sure it is publicly available.")     

        # Parse the URL to get the path
        path = urlparse(cloudFile.file_identifier).path

        # Extract the file extension
        file_extension = os.path.splitext(path)[1]

        # Download file
        with tempfile.NamedTemporaryFile(delete=False, suffix=file_extension) as temp_file:
            temp_file.write(response.content)
            yield LocalFile(file_path=temp_file.name, metadata=cloudFile.metadata, id=cloudFile.id, type=file_extension)

    def config_validation(self) -> bool:
        import requests
        try:
            response = requests.get(self.url)
            if not response.ok:
                raise NeumFileException(f"File can't be accessed. Please make sure it is publicly available.")     
        except Exception as e:
            raise NeumFileException(f"Connection to file failed, check url. See Exception: {e}")     
        # Check for metadata
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided") 
        return True 



from psycopg2.extras import DictCursor
from datetime import datetime
from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator, Optional
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.Shared.Selector import Selector
from neumai.Shared.Exceptions import PostgresConnectionException
from decimal import Decimal
from pydantic import Field
import psycopg2
import json

class PostgresConnector(DataConnector):
    """
    Postgres Connector

    Extracts data from any Postgres database using a given query.
    
    Attributes:
    -----------

    connection_string : str
        Connection string for the Postgres database.
    query : str
        Query to be executed to pull data from the database. (i.e. Select * From TableName)
    selector : Optional[Selector]
        Optional selector object to define what data data should be used to generate embeddings or stored as metadata with the vector.
    
    """

    connection_string: str = Field(..., description="Connection string for the Postgres database.")

    query: str = Field(..., description="Query to execute on the Postgres database.")

    batch_size: Optional[int] = Field(1000, description="Batch size for processing data.")
    
    selector: Optional[Selector] = Field(Selector(to_embed=[], to_metadata=[]), description="Selector for data connector metadata")

    @property
    def connector_name(self) -> str:
        return "PostgresConnector"

    @property
    def required_properties(self) -> List[str]:
        return ["connection_string", "query"]

    @property
    def optional_properties(self) -> List[str]:
        return ["batch_size"]
    
    @property
    def available_metadata(self) -> str:
        return []

    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["JSONLoader"]
    
    class CustomEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, Decimal):
                return str(obj)  # Convert Decimal to a string representation
            elif isinstance(obj, datetime):
                return obj.isoformat()  # Convert datetime to ISO format string
            return super().default(obj)
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        connection_string = self.connection_string
        query = self.query
        
        # Optional
        batch_size = self.batch_size

        with psycopg2.connect(connection_string) as connection:
            with connection.cursor(cursor_factory=DictCursor, name='neumai') as cursor:
                cursor.itersize = 1000  # fetch 1000 rows at a time
                cursor.execute(query)
                batch_rows = []
                for row in cursor:
                    serialized_string = json.dumps(dict(row),cls=self.CustomEncoder)
                    serialized_dict = json.loads(serialized_string)
                    batch_rows.append(serialized_dict)
                    if(len(batch_rows) == batch_size):
                        yield CloudFile(data=json.dumps(batch_rows), metadata={}, id="Postgres")
                        batch_rows = []
                if len(batch_rows) > 0:
                    yield CloudFile(data=json.dumps(batch_rows), metadata={}, id="Postgres")

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # No metadatadata to determine what rows are new. Needs to be done through websocket
        yield from self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        data = json.loads(cloudFile.data)
        for row in data:
            yield LocalFile(in_mem_data=json.dumps(row), metadata=cloudFile.metadata)

    def config_validation(self) -> bool:
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        try:
            psycopg2.connect(self.connection_string)
        except Exception as e:
            raise PostgresConnectionException(f"Connection to Postgres failed, check credentials. See Exception: {e}")      
        return True   


from .PipelineRun import PipelineRun
from .TriggerSchedule import TriggerSchedule
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from neumai.ModelFactories import EmbedConnectorFactory, SinkConnectorFactory
from neumai.Sources.SourceConnector import SourceConnector
from neumai.Shared.NeumVector import NeumVector
from neumai.Shared.NeumSearch import NeumSearchResult
from typing import List, Optional
from pydantic import BaseModel, Field, validator
from uuid import uuid4
import json

class Pipeline(BaseModel):
    """
    Pipeline managing the flow of data from sources through embedding to the sink, including configurations and state.

    This class orchestrates the movement and transformation of data, coordinating between source connectors, embedding processes, and sinks. It also handles pipeline configurations and maintains the state of the pipeline operations.

    Attributes:
    -----------
    id : str
        Unique identifier for the pipeline, generated using UUID4.
    
    sources : List[SourceConnector]
        List of source connectors involved in the pipeline for data input.

    embed : EmbedConnector
        Embedding connector for processing data.

    sink : SinkConnector
        Sink connector for the final output of processed data.

    name : Optional[str]
        Name of the pipeline, providing a human-readable identifier.

    created : Optional[float]
        Timestamp of when the pipeline was created, represented as a float.

    updated : Optional[float]
        Timestamp of when the pipeline was last updated.


    trigger_schedule : Optional[TriggerSchedule]
        Scheduling details for triggering the pipeline execution.

    latest_run : Optional[PipelineRun]
        Information about the latest run of the pipeline.

    owner : Optional[str]
        Owner of the pipeline, indicating who manages or created it.

    is_deleted : bool
        Flag to indicate if the pipeline is marked as deleted. Default is False.
    """

    id: str = Field(default_factory=lambda: str(uuid4()), description="Unique identifier for the pipeline")

    name: Optional[str] = Field(None, description="Name of the pipeline")

    description: Optional[str] = Field(None, description="Description of the data contained in the pipeline")

    created: Optional[float] = Field(None, description="Timestamp of when the pipeline was created")

    updated: Optional[float] = Field(None, description="Timestamp of when the pipeline was last updated")

    sources: List[SourceConnector] = Field(..., description="List of source connectors involved in the pipeline")

    embed: EmbedConnector = Field(..., description="Embedding connector for data processing")

    sink: SinkConnector = Field(..., description="Sink connector for final data output")

    trigger_schedule: Optional[TriggerSchedule] = Field(None, description="Scheduling details for triggering the pipeline")

    latest_run: Optional[PipelineRun] = Field(None, description="Information about the latest run of the pipeline")

    owner: Optional[str] = Field(None, description="Owner of the pipeline")
    
    is_deleted: bool = Field(False, description="Flag to indicate if the pipeline is marked as deleted")

    def set_id(self, id: str):
        self.id = id
    
    def set_created(self, created: float):
        self.created = created

    def set_updated(self, updated: float):
        self.updated = updated

    def set_latest_run(self, pipeline_run: PipelineRun):
        self.latest_run = pipeline_run

    def set_owner(self, owner: str):
        self.owner = owner
    
    def available_metadata(self) -> List[str]:
        available_metadata = []
        for source in self.sources:
            available_metadata += source.custom_metadata.keys()
        # Check if data_connector exists and has a selector
        if hasattr(source.data_connector, 'selector') and source.data_connector.selector:
            available_metadata += source.data_connector.selector.to_metadata

        # Check if loader exists and has a selector
        if hasattr(source.loader, 'selector') and source.loader.selector:
            available_metadata += source.loader.selector.to_metadata
        return available_metadata

    @validator("embed", pre=True, always=True)
    def deserialize_embed(cls, value):
        if isinstance(value, dict):
            return EmbedConnectorFactory.get_embed(value.get("embed_name"), value.get("embed_information"))
        return value
    
    @validator("sink", pre=True, always=True)
    def deserialize_sink(cls, value):
        if isinstance(value, dict):
            return SinkConnectorFactory.get_sink(value.get("sink_name"), value.get("sink_information"))
        return value
    
    def config_validation(self) -> bool:
        """Running validation for each connector"""
        try:
            for source in self.sources:
                source.validation()
            self.embed.validation()
            self.sink.validation()
            return True
        except Exception as e:
            raise e
    
    def run(self) -> int:
        # This method is meant for local development only. Not to be used in production.
        # The Neum AI framework provides parallelization constructs through yielding
        # These should be used to run pipelines at scale.
        try:
            self.config_validation()
        except Exception as e:
            raise e
        
        try:
            total_vectors_stored = 0
            for source in self.sources:
                for cloudFile in source.list_files_full():
                    for localFile in source.download_files(cloudFile=cloudFile):
                        for document in source.load_data(file=localFile):
                            for chunks in source.chunk_data(document=document):
                                embeddings, embeddings_info = self.embed.embed(documents=chunks)
                                vectors_to_store = [NeumVector(id=str(uuid4()), vector=embeddings[i], metadata=chunks[i].metadata) for i in range(0,len(embeddings))]
                                total_vectors_stored += self.sink.store(vectors_to_store=vectors_to_store)
            return total_vectors_stored
        except Exception as e:
            raise e
    
    def search(self, query:str, number_of_results:int, filter:dict={}) -> List[NeumSearchResult]:
        vector_for_query = self.embed.embed_query(query=query)
        matches =  self.sink.search(vector=vector_for_query, number_of_results=number_of_results, filter=filter)
        return matches

    # Todo standardize the model serialization as we are mixing FE and BE concepts into the SDK
    def as_pipeline_model(self):
        content_to_return = {}
        content_to_return['id'] = self.id
        content_to_return['name'] = self.name
        content = []
        for source in self.sources:
            content.append(source.as_json())
        content_to_return['sources'] = content
        content_to_return['embed'] = self.embed.as_json()
        content_to_return['sink'] = self.sink.as_json()
        content_to_return['created'] = self.created
        content_to_return['updated'] = self.updated
        if self.trigger_schedule == None:
            content_to_return['trigger_schedule'] = None
        else:
            content_to_return['trigger_schedule'] = json.loads(self.trigger_schedule.json())
        if self.latest_run == None:
            content_to_return['latest_run'] = None
        else:
            content_to_return['latest_run'] = json.loads(self.latest_run.json())
        content_to_return['available_metadata'] = self.available_metadata()
        content_to_return['is_deleted'] = self.is_deleted
        content_to_return['owner'] = self.owner

        return content_to_return

    def as_json(self):
        json_body = {}
        json_body['name'] = self.name
        json_source = []
        for source in self.sources:
            json_source.append(source.as_json())
        json_body['sources'] = json_source
        json_body['embed'] = self.embed.as_json()
        json_body['sink'] = self.sink.as_json()
        if self.trigger_schedule == None:
            json_body['trigger_schedule'] = None
        else:
            json_body['trigger_schedule'] = json.loads(self.trigger_schedule.json())
        return json_body


from .Loader import Loader
from .AutoLoader import AutoLoader
from .HTMLLoader import HTMLLoader
from .MarkdownLoader import MarkdownLoader
from .PDFLoader import PDFLoader
from .CSVLoader import CSVLoader
from .JSONLoader import JSONLoader
from .LoaderEnum import LoaderEnum


from .DataConnector import DataConnector
from .AzureBlobConnector import AzureBlobConnector
from .WebsiteConnector import WebsiteConnector
from .S3Connector import S3Connector
from .SupabaseStorageConnector import SupabaseStorageConnector
from .PostgresConnector import PostgresConnector
from .SharepointConnector import SharepointConnector
from .SingleStoreConnector import SingleStoreConnector
from .FileConnector import FileConnector
from .DataConnectorEnum import DataConnectorEnum


from typing import Any, Optional
from .TriggerSyncTypeEnum import TriggerSyncTypeEnum
from pydantic import BaseModel

class PipelineRunTaskDetails(BaseModel):

    completed_embedding_tasks: int
    completed_storing_tasks: int
    failed_embedding_tasks: int
    failed_storing_tasks: int

    def toJson(self):
        json_to_return = {}
        json_to_return['completed_embedding_tasks'] = self.completed_embedding_tasks
        json_to_return['completed_storing_tasks'] = self.completed_storing_tasks
        json_to_return['failed_embedding_tasks'] = self.failed_embedding_tasks
        json_to_return['failed_storing_tasks'] = self.failed_storing_tasks

        return json_to_return
    
    def as_pipeline_run_task_details(task_details: dict):
        if task_details == None:
            return None
        return PipelineRunStatus(
            completed_embedding_tasks=task_details.get("completed_embedding_tasks", None),
            completed_storing_tasks=task_details.get("completed_storing_tasks", None),
            failed_embedding_tasks=task_details.get("failed_embedding_tasks", None),
            failed_storing_tasks=task_details.get("failed_storing_tasks", None),
        )
    
class PipelineRunStatus(BaseModel):
    status: str
    message: Optional[str] = None
    exception_detail: Optional[str] = None

    def __init__(self, **data):
        super().__init__(**data)
        self.status = self.status.lower()
    
    def as_pipeline_run_status(detailed_status: dict):
        if detailed_status == None:
            return PipelineRunStatus(status="Unknown",message="Something went wrong in our end, please email kevin@tryneum.com and we will get you unblocked")
        return PipelineRunStatus(
            status=detailed_status.get("status", None),
            message=detailed_status.get("message", None),
            exception_detail=detailed_status.get("exception_detail", None),
        )
    
class PipelineRun(BaseModel):
    id: str
    created: float
    pipeline_id: str
    trigger_type: str
    sync_type: TriggerSyncTypeEnum
    detailed_status: Optional[PipelineRunStatus] = None
    vectors_written: int = 0
    task_details: Optional[PipelineRunTaskDetails] = None
    last_updated: Optional[float] = None
    number_of_documents: Optional[int] = None
    finished_distributing: bool = False
    processing_time: Optional[float] = None
                                                 
    def set_id(self, id: str):
        self.id = id

    def set_number_of_documents(self, number_of_documents):
        self.number_of_documents = number_of_documents
    
    def set_finished_distributing(self, finished_distributing):
        self.finished_distributing = finished_distributing

    def set_detailed_status(self, detailed_status: PipelineRunStatus):
        self.detailed_status = detailed_status
    
    def set_pipeline_id(self, pipeline_id: str):
        self.pipeline_id = pipeline_id

    def set_created(self, created: float):
        self.created = created

    def set_processing_time(self, processing_time: float):
        self.processing_time = processing_time
    
    def as_pipeline_run(dct:dict):
        if dct == None:
            return PipelineRun(id=None, created=None, pipeline_id = None, trigger_type=None, sync_type=None, detailed_status=None, vectors_written=None, number_of_documents=None, finished_distributing=None)
        return PipelineRun(
            id=dct.get("id", None),
            pipeline_id=dct.get("pipeline_id"),
            created=dct.get("created", None),
            trigger_type=dct.get("trigger_type", None),
            sync_type=TriggerSyncTypeEnum.as_trigger_sync_type(trigger_sync_type=dct.get("sync_type", None)),
            detailed_status=PipelineRunStatus.as_pipeline_run_status(detailed_status=dct.get("detailed_status", None)),
            vectors_written=dct.get("vectors_written", None),
            task_details=PipelineRunTaskDetails.as_pipeline_run_task_details(dct.get("task_details", None)),
            last_updated=dct.get("last_updated", None),
            number_of_documents=dct.get("number_of_documents",None),
            finished_distributing=dct.get("finished_distributing",None),# we could do something like in distributed tasks where we store the state of the DAG in the pipeline run object.. for now just doing finished_distributing
            processing_time=dct.get("processing_time", None)
        )


from typing import List, Tuple, Optional
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from langchain.embeddings.openai import OpenAIEmbeddings
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.Exceptions import OpenAIConnectionException
from pydantic import BaseModel, Field

class OpenAIEmbed(EmbedConnector):
    """
    OpenAI Embed Connector

    Facilitates embedding and processing data using OpenAI's services. This connector is designed to interact directly with OpenAI's APIs, requiring an API key for authentication and operation.

    Attributes:
    -----------
    api_key : str
        The API key for accessing OpenAI services. Essential for authenticating and using OpenAI's API for data processing.

    max_retries : Optional[int]
        The maximum number of retries for connection attempts with OpenAI's services. Default is 20 retries.

    chunk_size : Optional[int]
        The size of chunks for processing data, defined by the number of items. Default is 1000, suitable for large data sets.
    """

    api_key: str = Field(..., description="API key for OpenAI services.")
    
    max_retries: Optional[int] = Field(20, description="Maximum number of retries for the connection.")

    chunk_size: Optional[int] = Field(1000, description="Size of chunks for processing data.")

    @property
    def embed_name(self) -> str:
        return 'OpenAIEmbed'
    
    @property
    def required_properties(self) -> List[str]:
        return ["api_key"]

    @property
    def optional_properties(self) -> List[str]:
        return ['max_retries', 'chunk_size']

    def validation(self) -> bool:
        """config_validation connector setup"""
        try:
            OpenAIEmbeddings(max_retries=20, api_key=self.api_key, chunk_size=1000)
        except Exception as e:
            raise OpenAIConnectionException(f"OpenAI couldn't be initialized. See exception: {e}")
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with OpenAI"""

        embedding = OpenAIEmbeddings(
            max_retries=self.max_retries,
            chunk_size=self.chunk_size,
            api_key=self.api_key, 
        )
        embeddings = []
        texts = [x.content for x in documents]
        # do we want to persist some embeddings if they were able to be wrriten but not another "batch" of them? or should we treat all texts as an atomic operation
        embeddings  = embedding.embed_documents(texts=texts)
        #cost_per_token = 0.000000001 # ADA-002 as of Sept 2023
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }
        return embeddings,info

    def embed_query(self, query: str) -> List[float]:
        """Generate embeddings for a single query using OpenAI"""
        embedding = OpenAIEmbeddings(api_key=self.api_key)
        return embedding.embed_query(query)



from abc import abstractmethod, ABC
from typing import List, Generator
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.Shared.Selector import Selector
from datetime import datetime
from pydantic import BaseModel
import json

class DataConnector(BaseModel, ABC):

    @property
    @abstractmethod
    def connector_name(self) -> str:
        pass

    @property
    @abstractmethod
    def required_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optional_properties(self) -> List[str]:
        pass
    
    @property
    @abstractmethod
    def available_metadata(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def schedule_avaialable(self) -> bool:
        pass

    @property
    @abstractmethod
    def auto_sync_available(self) -> bool:
        pass

    @property
    @abstractmethod
    def compatible_loaders(self) -> List[str]:
        pass

    @abstractmethod
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        """Connect to source and download file into local storage"""
    
    @abstractmethod
    def connect_and_list_delta(self, last_run:datetime) -> Generator[LocalFile, None, None]:
        """Check for changes in the source"""
        """Code to be pushed to a worker and run on a schedule"""
    
    @abstractmethod
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        """Connect to source and download file into local storage"""

    @abstractmethod
    def config_validation(self) -> bool:
        """config_validation if the connector is correctly configured"""

    # To do auto_sync logic.

    def as_json(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['connector_name'] = self.connector_name
        json_to_return['connector_information'] = json.loads(self.json())
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['required_properties'] = self.required_properties
        json_to_return['optional_properties'] = self.optional_properties
        json_to_return['available_metadata'] = self.available_metadata
        json_to_return['compatible_loaders'] = self.compatible_loaders
        json_to_return['schedule_avaialable'] = self.schedule_avaialable
        json_to_return['auto_sync_available'] = self.auto_sync_available
        return json_to_return


from typing import List, Generator, Dict, Optional
from datetime import datetime
from pydantic import BaseModel, Field, validator
from neumai.DataConnectors.DataConnector import DataConnector
from neumai.Chunkers.Chunker import Chunker
from neumai.Chunkers.RecursiveChunker import RecursiveChunker
from neumai.Loaders.Loader import Loader
from neumai.Loaders.AutoLoader import AutoLoader
from neumai.ModelFactories import ChunkerFactory, DataConnectorFactory, LoaderFactory
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile

# Assuming DataConnector, Chunker, RecursiveChunker, Loader, and AutoLoader are defined elsewhere
class SourceConnector(BaseModel):
    """
    Source Connector

    This class acts as a constructor for source connectors, encompassing various components like data connectors, chunkers, and loaders. It facilitates the initial stages of data processing by managing the connection to data sources, breaking down content, and loading data for further processing.

    Attributes:
    -----------
    data_connector : DataConnector
        A connector to the data source. This component is responsible for establishing and managing the connection to the source of the data.

    chunker : Chunker
        The chunker to be used for breaking down the content. It determines how the data is segmented for processing. Default is set to RecursiveChunker, which recursively breaks down the content based on specified rules.

    loader : Loader
        The loader responsible for loading data from files or specific data types. It handles the retrieval and initial processing of data from its source. Default is set to AutoLoader, which automatically determines the appropriate method for loading data.

    custom_metadata : Dict
        A dictionary for custom metadata that can be added to the vector. This allows for the inclusion of additional, user-defined information in the data processing pipeline.
    """

    data_connector: DataConnector = Field(..., description="Connector to data source")

    chunker: Chunker = Field(default=RecursiveChunker(), description="Chunker to be used to break down content")

    loader: Loader = Field(default=AutoLoader(), description="Loader to load data from file / data type")

    custom_metadata: Optional[Dict] = Field(default_factory=dict, description="Custom metadata to be added to the vector")

    def list_files_full(self) -> Generator[CloudFile, None, None]:
        yield from self.data_connector.connect_and_list_full()

    def list_files_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        yield from self.data_connector.connect_and_list_delta(last_run=last_run)

    def download_files(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        yield from self.data_connector.connect_and_download(cloudFile=cloudFile)

    def load_data(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        yield from self.loader.load(file=file)

    def chunk_data(self, document:NeumDocument) -> Generator[List[NeumDocument], None, None]:
        for chunk_set in self.chunker.chunk(documents=[document]):
            chunk_set_with_custom_metadata = [NeumDocument(id=chunk.id, content=chunk.content, metadata={**chunk.metadata, **self.custom_metadata, **{"text":chunk.content}}) for chunk in chunk_set]
            yield chunk_set_with_custom_metadata
    
    def validation(self) -> bool:
        core_validation = self.data_connector.config_validation() and self.loader.config_validation() and self.chunker.config_validation()
        loader_validation = self.loader.loader_name in self.data_connector.compatible_loaders
        return core_validation and loader_validation

    @validator("data_connector", pre=True, always=True)
    def deserialize_data_connector(cls, value):
        if isinstance(value, dict):
            return DataConnectorFactory.get_data_connector(value.get("connector_name"), value.get("connector_information"))
        return value
    
    @validator("chunker", pre=True, always=True)
    def deserialize_chunker(cls, value):
        if isinstance(value, dict):
            return ChunkerFactory.get_chunker(value.get("chunker_name"), value.get("chunker_information"))
        return value
        
    @validator("loader", pre=True, always=True)
    def deserialize_loader(cls, value):
        if isinstance(value, dict):
            return LoaderFactory.get_loader(value.get("loader_name"), value.get("loader_information"))
        return value
    
    def as_json(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['custom_metadata'] = self.custom_metadata
        json_to_return['data_connector'] = self.data_connector.as_json()
        json_to_return['chunker'] = self.chunker.as_json()
        json_to_return['loader'] = self.loader.as_json()
        return json_to_return



from datetime import datetime
from neumai.DataConnectors import DataConnector
from typing import List, Generator, Optional
from azure.storage.blob import BlobClient, ContainerClient
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.Shared.Exceptions import AzureBlobConnectionException
from neumai.Shared.Selector import Selector
import tempfile
import os
from pydantic import Field

class AzureBlobConnector(DataConnector):
    """
    Azure Blob data connector

    Extracts data from a Azure Blob container. 
    
    Attributes:
    -----------

    connection_string : str
        Connection string to the Azure Blob
    container_name : str
        Name of the Azure Blob container you want to extract data from
    selector : Optional[Selector]
        Optional selector object to define what data data should be used to generate embeddings or stored as metadata with the vector.
    
    """

    connection_string: str = Field(..., description="Connection string to connect to Azure Blob [required]")

    container_name: str = Field(..., description="Container name to connect to [required]")

    selector: Optional[Selector] = Field(Selector(to_embed=[], to_metadata=[]), description="Selector for data connector metadata")

    @property
    def connector_name(self) -> str:
        return "AzureBlobConnector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["connection_string", "container_name"]

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> str:
        return ['name', 'last_modified', 'creation_time', 'last_access_on']

    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "CSVLoader", "JSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        container = ContainerClient.from_connection_string(
            conn_str=self.connection_string, container_name=self.container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            name = file.name
            metadata = {
                "creation_time": file.creation_time.isoformat(),
                "last_modified": file.last_modified.isoformat(),
                "last_access_on": file.last_accessed_on.isoformat() if file.last_accessed_on is not None else None
            }
            selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
            yield CloudFile(file_identifier=name, metadata=selected_metadata, id = name)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        container = ContainerClient.from_connection_string(
            conn_str=self.connection_string, container_name=self.container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            last_update_date = file.last_modified
            if(last_run < last_update_date):
                name = file.name
                metadata = {
                    "creation_time": file.creation_time.isoformat(),
                    "last_modified": file.last_modified.isoformat(),
                    "last_access_on": file.last_accessed_on.isoformat() if file.last_accessed_on is not None else None
                }
                selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
                yield CloudFile(file_identifier=name, metadata=selected_metadata, id=name)

    def connect_and_download(self,  cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        client = BlobClient.from_connection_string(conn_str=self.connection_string, container_name=self.container_name, blob_name=cloudFile.file_identifier)
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{self.container_name}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(f"{file_path}", "wb") as file:
                blob_data = client.download_blob()
                blob_data.readinto(file)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)
        
    def config_validation(self) -> bool:
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")

        try:
            ContainerClient.from_connection_string(
                conn_str=self.connection_string, container_name=self.container_name
            )
        except Exception as e:
            raise AzureBlobConnectionException(f"Connection to Azure Blob Storage failed, check credentials. See Exception: {e}")
        return True     


    
 


from neumai.EmbedConnectors import (
    EmbedConnector,
    OpenAIEmbed,
    ReplicateEmbed,
    AzureOpenAIEmbed
)

from neumai.EmbedConnectors.EmbedConnectorEnum import EmbedConnectorEnum
from neumai.Shared.Exceptions import InvalidEmbedConnectorException


available_embed_connectors = [enum.value for enum in list(EmbedConnectorEnum)]

class EmbedConnectorFactory:
    """Class that leverages the Factory pattern to get the appropriate embed connector
    """
    def get_embed(embed_name: str, embed_information: dict) ->EmbedConnector:
        embed_connector_name = embed_name.replace(" ","").lower()
        embed_connector_enum = EmbedConnectorEnum.as_embed_connector_enum(embed_connector_name=embed_connector_name)
        if embed_connector_enum == EmbedConnectorEnum.azureopenaiembed:
            return AzureOpenAIEmbed(**embed_information)
        elif embed_connector_enum == EmbedConnectorEnum.openaiembed:
            return OpenAIEmbed(**embed_information)
        elif embed_connector_enum == EmbedConnectorEnum.replicateembed:
            return ReplicateEmbed(**embed_information)
        else:
            raise InvalidEmbedConnectorException(f"{embed_connector_name} is an invalid embed connector. Available connectors: {available_embed_connectors}] ")


from typing import List, Optional, Tuple
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.Shared.NeumVector  import NeumVector
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from neumai.Shared.NeumSearch import NeumSearchResult
from neumai.Shared.Exceptions import(
    WeaviateConnectionException,
    WeaviateInsertionException,
    WeaviateIndexInfoException,
    WeaviateQueryException
)
from pydantic import Field
from weaviate.util import generate_uuid5, _capitalize_first_letter
import weaviate

class WeaviateSink(SinkConnector):
    """
    Weaviate Sink

    A sink connector specifically designed for integrating with Weaviate, an open-source, ML-powered search engine. This connector enables efficient data output to a Weaviate instance, supporting various configuration options to optimize the data transfer process.

    Attributes:
    -----------
    url : str
        The URL for accessing the Weaviate instance. This should be the endpoint where Weaviate is hosted.

    api_key : str
        The API key for authenticating with the Weaviate service. This key is necessary for secure communication with the Weaviate instance.

    class_name : str
        An optional class name within Weaviate to which the data will be associated. Specifies the schema or type of data being stored.

    num_workers : Optional[int]
        The optional number of worker threads to use for data processing and uploading. Default is 1.

    shard_count : Optional[int]
        The optional number of shards to use for distributing data within Weaviate. Default is 1.

    batch_size : Optional[int]
        The optional size of batches for data processing and upload. Defines how many data items are processed together. Default is 100.

    is_dynamic_batch : Optional[bool]
        An optional flag to enable dynamic batching. If set to True, batch sizes may be adjusted dynamically based on network conditions and load. Default is False.

    batch_connection_error_retries : Optional[int]
        The optional number of retries in case of a batch connection error. Specifies how many times the connector should attempt to resend a batch in case of a connection error. Default is 3 retries.
    """

    url: str = Field(..., description="URL for Weaviate.")

    api_key: str = Field(..., description="API key for Weaviate.")

    class_name: str = Field(..., description="Class name.")

    num_workers: Optional[int] = Field(1, description="Optional number of workers.")

    shard_count: Optional[int] = Field(1, description="Optional shard count.")

    batch_size: Optional[int] = Field(100, description="Optional batch size.")

    is_dynamic_batch: Optional[bool] = Field(False, description="Optional dynamic batch flag.")

    batch_connection_error_retries: Optional[int] = Field(3, description="Optional batch connection error retries.")

    @property
    def sink_name(self) -> str:
        return 'WeaviateSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['url', 'api_key', 'class_name']

    @property
    def optional_properties(self) -> List[str]:
        return ['num_workers', 'shard_count', 'batch_size', 'is_dynamic_batch', 'batch_connection_error_retries']

    def validation(self) -> bool:
        """config_validation connector setup"""
        try:
            if 'https' not in self.url:
                client = weaviate.Client(
                    url=self.url
                )
            else:
                api_key = self.api_key
                client = weaviate.Client(
                    url=self.url,
                    auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
                )
        except Exception as e:
            raise WeaviateConnectionException(f"Weaviate couldn't be initialized. See exception: {e}")
        return True 

    def _check_batch_result(self, results: Optional[List[dict[str, any]]], partial_failure: dict):
        if results is not None:
            for result in results:
                if "result" in result and "errors" in result["result"]:
                    if "error" in result["result"]["errors"]:
                        print(f"[ERROR] Error when batching to weaviate {result['result']}")
                        partial_failure['did_fail'] = True
                        partial_failure['latest_failure'] = result["result"]["errors"]["error"]
                        partial_failure['number_of_failures'] += 1

    def delete_vectors_with_file_id(self, file_id: str) -> bool:
        api_key = self.api_key
        url = self.url
        # Weaviate requires first letter to be capitalized
        class_name = self.class_name.replace("-","_")
        class_name = _capitalize_first_letter(class_name)
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        client.batch.delete_objects(
            class_name=class_name,
            where={
                "path": ["_file_entry_id"],
                "operator": "Equal",
                "valueText": file_id
            },
        )
        return True
    
    def store(self, vectors_to_store:List[NeumVector]) -> Tuple[List, dict]:
        url = self.url
        num_workers = self.num_workers
        shard_count = self.shard_count
        batch_size = self.batch_size
        is_dynamic_batch = self.is_dynamic_batch
        batch_connection_error_retries = self.batch_connection_error_retries
        class_name = self.class_name.replace("-","_")
        class_name = _capitalize_first_letter(class_name)
        partial_failure = {'did_fail': False, 'latest_failure': None, 'number_of_failures': 0}

        if 'https' not in url:
            client = weaviate.Client(
                url=url
            )
        else:
            api_key = self.api_key
            client = weaviate.Client(
                url=url,
                auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
            )
        try:
            client.schema.create_class({
                "class": class_name,
                "shardingConfig":{"desiredCount":shard_count},
            })
        except weaviate.UnexpectedStatusCodeException as e:
            if 'already exists' not in e.message:
                raise WeaviateInsertionException(f"Error when creating class in weaviate. Error: {e}")
            
        with client.batch.configure(
            batch_size=batch_size,
            callback=lambda results: self._check_batch_result(results, partial_failure),
            num_workers=num_workers,
            dynamic=is_dynamic_batch,
            connection_error_retries=batch_connection_error_retries
        ) as batch:
            for i in range(0, len(vectors_to_store)):
                try:
                    batch.add_data_object(
                        data_object=vectors_to_store[i].metadata,
                        class_name=class_name,
                        vector=vectors_to_store[i].vector,
                        uuid=generate_uuid5(vectors_to_store[i].id)
                    )
                except Exception as e:
                    raise WeaviateInsertionException(f"Error when adding data object to Weaviate. Error: {str(e)}")

        return len(vectors_to_store)

    def search(self, vector: List[float], number_of_results: int, filter:dict={}) -> List[NeumSearchResult]:
        api_key = self.api_key
        url = self.url
        # Weaviate requires first letter to be capitalized
        class_name = self.class_name.replace("-","_")
        class_name = _capitalize_first_letter(class_name)
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        
        try:
            class_schema = client.schema.get(class_name)
        except Exception as e:
            raise WeaviateQueryException(f"There was an error retrieving the class schema from weaviate")

        full_class_schema_properties = [property['name'] for property in class_schema['properties']]
        matches = []
        try:
            client_query = (
                client.query
                .get(class_name=class_name, properties=full_class_schema_properties)
                .with_near_vector(content={'vector': vector})
                .with_limit(number_of_results)
                .with_additional(['id', 'certainty', 'vector'])
            )

            # Add .with_where(filter) only if filter is not empty
            if filter:
                client_query = client_query.with_where(filter)

            # Final execution of the query
            search_result = client_query.do()

            for result in search_result["data"]["Get"][class_name]:
                # unify our api with the metadata.. or just return whatever metadata we have. (?)
                matches.append(NeumSearchResult(id=result['_additional']['id'], score=result['_additional']['certainty'], metadata= {k: v for k, v in result.items() if k != "_additional"}, vector=result['_additional']['vector']))
        except Exception as e:
            raise WeaviateQueryException(f"There was an error querying weaviate. Error {e}")
        return matches

    def info(self) -> NeumSinkInfo:
        api_key = self.api_key
        url = self.url
        
        class_name = self.class_name.replace("-","_")
        class_name = _capitalize_first_letter(class_name)
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        try:
            response = (
                client.query
                .aggregate(class_name=class_name)
                .with_meta_count()
                .do()
            )
            vectors_stored_in_class = response["data"]["Aggregate"][class_name]["meta"]["count"]
            return NeumSinkInfo(number_vectors_stored=vectors_stored_in_class)
        except Exception as e:
            raise WeaviateIndexInfoException(f"There was an error getting class info from weaviate {e}")


from abc import abstractmethod, ABC
from typing import List, Generator
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.Selector import Selector
from pydantic import BaseModel
import json

class Loader(ABC, BaseModel):    
    @property
    @abstractmethod
    def loader_name(self) -> str:
        pass
    
    @property
    @abstractmethod
    def required_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optional_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def available_metadata(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def available_content(self) -> List[str]:
        pass

    @abstractmethod
    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""

    @abstractmethod
    def config_validation(self) -> bool:
        """config_validation if the loader is correctly configured"""

    def as_json(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['loader_name'] = self.loader_name
        json_to_return['loader_information'] = json.loads(self.json())
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['required_properties'] = self.required_properties
        json_to_return['optional_properties'] = self.optional_properties
        json_to_return['available_metadata'] = self.available_metadata
        json_to_return['available_content'] = self.available_content
        return json_to_return


from typing import List, Tuple, Optional
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from neumai.Shared.NeumDocument import NeumDocument
from langchain.embeddings import azure_openai
from pydantic import Field

class AzureOpenAIEmbed(EmbedConnector):
    """
    Azure OpenAI Embed Connector

    This class is designed for integrating with Azure's deployment of OpenAI, allowing for embedding and processing data using OpenAI's models hosted on Azure. It handles connections and data processing tasks with Azure OpenAI services.

    Attributes:
    -----------
    deployment_name : str
        The name of the Azure OpenAI deployment. Used to specify and connect to the particular Azure OpenAI service.

    api_key : str
        The API key for accessing the Azure OpenAI services. This key is necessary for authentication and usage of the services.

    endpoint : str
        The specific endpoint URL for the Azure OpenAI service. This is the entry point for sending requests to the Azure OpenAI API.

    max_retries : Optional[int]
        The maximum number of retries for the connection to the Azure OpenAI service. Default is 20 retries.

    chunk_size : Optional[int]
        The size of chunks for processing data, specified in terms of number of items. Default is 16.
    """

    deployment_name: str = Field(..., description="Deployment name for Azure OpenAI.")

    api_key: str = Field(..., description="API key for Azure OpenAI.")

    endpoint: str = Field(..., description="Endpoint for Azure OpenAI.")

    max_retries: Optional[int] = Field(20, description="Maximum number of retries for the connection.")

    chunk_size: Optional[int] = Field(16, description="Size of chunks for processing data.")

    @property
    def embed_name(self) -> str:
        return 'AzureOpenAIEmbed'

    @property
    def required_properties(self) -> List[str]:
        return ['deployment_name', 'api_key', 'endpoint']

    @property
    def optional_properties(self) -> List[str]:
        return ['max_retries', 'chunk_size']
    
    def validation(self) -> bool:
        """config_validation connector setup"""
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with Azure OpenAI"""
        embedding = azure_openai.AzureOpenAIEmbeddings(
            max_retries=self.max_retries,
            chunk_size=max(self.chunk_size, 16),
            azure_deployment=self.deployment_name,
            api_key=self.api_key,
            azure_endpoint=self.endpoint
        )
        embeddings = []
        texts = [x.content for x in documents]
        embeddings  = embedding.embed_documents(texts=texts)
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }
        return embeddings,info
    
    def embed_query(self, query: str) -> List[float]:
        """Generate embeddings for a single query using Azure OpenAI"""

        embedding = azure_openai.AzureOpenAIEmbeddings(
            azure_deployment=self.deployment_name,
            api_key=self.api_key,
            azure_endpoint=self.endpoint
        )
        return embedding.embed_query(query)


from .ChunkerFactory import ChunkerFactory
from .DataConnectorFactory import DataConnectorFactory
from .EmbedConnectorFactory import EmbedConnectorFactory
from .LoaderFactory import LoaderFactory
from .SinkConnectorFactory import SinkConnectorFactory



from typing import List, Optional
from neumai.Shared.NeumSearch import NeumSearchResult
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.Shared.NeumVector  import NeumVector
from neumai.Shared.Exceptions import (
    PineconeConnectionException,
    PineconeInsertionException,
    PineconeIndexInfoException,
    PineconeQueryException,
)
from pydantic import Field
import pinecone

class  PineconeSink(SinkConnector):
    """
    Pinecone Sink

    A sink connector specifically designed for Pinecone, facilitating the output of processed data into a Pinecone environment.

    Attributes:
    -----------
    api_key : str
        API key for accessing the Pinecone service.

    environment : str
        The specific Pinecone environment to connect to.

    index : str
        The index in Pinecone where the data will be stored.

    namespace : str
        Namespace within the Pinecone environment. Used for organizing data.
    """

    api_key: str = Field(..., description="API key for Pinecone.")

    environment: str = Field(..., description="Pinecone environment.")

    index: str = Field(..., description="Index for Pinecone.")

    namespace: str = Field(..., description="Data namespace.")

    @property
    def sink_name(self) -> str:
        return 'PineconeSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['api_key', 'environment', 'index', 'namespace']

    @property
    def optional_properties(self) -> List[str]:
        return []

    def validation(self) -> bool:
        """config_validation connector setup"""
        import pinecone
        try:
            pinecone.init(api_key=self.api_key, environment=self.environment)    
            index = pinecone.Index(index_name=self.index)
            index.describe_index_stats()
        except Exception as e:
            raise PineconeConnectionException(f"Pinecone connection couldn't be initialized. See exception: {e}")
        return True 

    def delete_vectors_with_file_id(self, file_id: str) -> bool:
        api_key =  self.api_key
        environment = self.environment
        index = self.index
        namespace = self.namespace
        if environment == "gcp-starter":
            raise Exception("Pinecone does not support deleting vectors by metadata in the gcp starter environment")
        pinecone.init(      
            api_key=api_key,      
            environment=environment)    
        index = pinecone.Index(index)
        index.delete(filter={"_file_entry_id": {"$eq": file_id}}, namespace=namespace)
        return True

    def store(self, vectors_to_store:List[NeumVector]) -> int:
        api_key =  self.api_key
        environment = self.environment
        index = self.index
        namespace = self.namespace
        if environment == "gcp-starter": namespace = None # short-term fix given gcp-starter limitation

        try:
            pinecone.init(api_key=api_key, environment=environment)    
            index = pinecone.Index(index_name=index)
            batch_size = 32 # how does changing this affect ?
            vectors_stored = 0
            for i in range(0, len(vectors_to_store), batch_size):
                # set end position of batch
                i_end = min(i + batch_size, len(vectors_to_store))
                # get batch of texts and ids
                vector_batch = vectors_to_store[i:i_end]
                to_upsert = [(vector.id, vector.vector, vector.metadata) for vector in vector_batch]
                result = index.upsert(vectors=to_upsert, namespace=namespace)
                vectors_stored += result['upserted_count'] 
        except Exception as e:
            raise PineconeInsertionException(f"Failed to store in Pinecone. Exception - {e}")
        return int(vectors_stored)
    
    def search(self, vector: List[float], number_of_results:int, filter:dict = {}) -> List[NeumSearchResult]:
        import pinecone
        api_key =  self.api_key
        environment = self.environment
        index = self.index
        namespace = self.namespace
        if environment == "gcp-starter": namespace = None # short-term fix given gcp-starter limitation
        try:
            pinecone.init(      
                api_key=api_key,      
                environment=environment)    
            index = pinecone.Index(index)
            results = index.query(
                vector=vector, 
                filter=filter,
                top_k=number_of_results, 
                namespace=namespace, 
                include_values=False, 
                include_metadata=True)["matches"]
        except Exception as e:
            raise PineconeQueryException(f"Failed to query pinecone. Exception - {e}")
        
        matches = []
        for result in results:
            matches.append(NeumSearchResult(id= result["id"], metadata=result["metadata"], score=result["score"]))
        return matches
    
    def info(self) -> NeumSinkInfo:
        import pinecone
        api_key =  self.api_key
        environment = self.environment
        index = self.index
        namespace = self.namespace
        if environment == "gcp-starter": namespace = None # short-term fix given gcp-starter limitation
        
        try:
            pinecone.init(      
                api_key=api_key,      
                environment=environment)    
            index = pinecone.Index(index)
            namespaces = index.describe_index_stats()["namespace"]
            if namespace in namespaces:
                return NeumSinkInfo(number_vectors_stored=namespaces[namespace]["vector_count"])
        except Exception as e:
            raise PineconeIndexInfoException(f"Failed to get info for pinecone. Exception - {e}")